@@ -1,5 +1,5 @@
 // elaboration functions
-// Copyright (C) 2005, 2006 Red Hat Inc.
+// Copyright (C) 2005-2007 Red Hat Inc.
 //
 // This file is part of systemtap, and is free software.  You can
 // redistribute it and/or modify it under the terms of the GNU General
@@ -11,6 +11,7 @@
 #include "parse.h"
 #include "tapsets.h"
 #include "session.h"
+#include "util.h"
 
 extern "C" {
 #include <sys/utsname.h>
@@ -30,17 +31,6 @@ extern "C" {
 using namespace std;
 
 
-template <typename OUT, typename IN> inline OUT
-lex_cast(IN const & in)
-{
-  stringstream ss;
-  OUT out;
-  if (!(ss << in && ss >> out))
-    throw runtime_error("bad lexical cast");
-  return out;
-}
-
-
 // ------------------------------------------------------------------------
 
 
@@ -51,6 +41,7 @@ derived_probe::derived_probe (probe *p):
     {
       this->locations = p->locations;  
       this->tok = p->tok;
+      this->privileged = p->privileged;
       this->body = deep_copy_visitor::deep_copy(p->body);
     }
 }
@@ -65,69 +56,58 @@ derived_probe::derived_probe (probe *p, probe_point *l):
   if (p)
     {
       this->tok = p->tok;
+      this->privileged = p->privileged;
       this->body = deep_copy_visitor::deep_copy(p->body);
     }
 }
 
 
-// ------------------------------------------------------------------------
-// Members of derived_probe_group
+void
+derived_probe::printsig (ostream& o) const
+{
+  probe::printsig (o);
+  printsig_nested (o);
+}
 
 void
-derived_probe_group::register_probe(be_derived_probe* p)
+derived_probe::printsig_nested (ostream& o) const
 {
-  throw semantic_error ("unexpected registration of a be_derived_probe");
+  // We'd like to enclose the probe derivation chain in a /* */
+  // comment delimiter.  But just printing /* base->printsig() */ is
+  // not enough, since base might itself be a derived_probe.  So we,
+  // er, "cleverly" encode our nesting state as a formatting flag for
+  // the ostream.
+  ios::fmtflags f = o.flags (ios::internal);
+  if (f & ios::internal)
+    {
+      // already nested
+      o << " <- ";
+      base->printsig (o);
+    }
+  else
+    {
+      // outermost nesting
+      o << " /* <- ";
+      base->printsig (o);
+      o << " */";
+    }
+  // restore flags
+  (void) o.flags (f);
 }
 
 
-void
-derived_probe_group::register_probe(dwarf_derived_probe* p)
+probe_point*
+derived_probe::sole_location () const
 {
-  throw semantic_error ("unexpected registration of a dwarf_derived_probe");
+  if (locations.size() == 0)
+    throw semantic_error ("derived_probe with no locations", this->tok);
+  else if (locations.size() > 1)
+    throw semantic_error ("derived_probe with too many locations", this->tok);
+  else 
+    return locations[0];
 }
 
 
-void
-derived_probe_group::register_probe(hrtimer_derived_probe* p)
-{
-  throw semantic_error ("unexpected registration of a hrtimer_derived_probe");
-}
-
-
-void
-derived_probe_group::register_probe(mark_derived_probe* p)
-{
-  throw semantic_error ("unexpected registration of a mark_derived_probe");
-}
-
-
-void
-derived_probe_group::register_probe(never_derived_probe* p)
-{
-  throw semantic_error ("unexpected registration of a never_derived_probe");
-}
-
-
-void
-derived_probe_group::register_probe(profile_derived_probe* p)
-{
-  throw semantic_error ("unexpected registration of a profile_derived_probe");
-}
-
-
-void
-derived_probe_group::register_probe(timer_derived_probe* p)
-{
-  throw semantic_error ("unexpected registration of a timer_derived_probe");
-}
-
-
-void
-derived_probe_group::register_probe(perfmon_derived_probe* p)
-{
-  throw semantic_error ("unexpected registration of a perfmon_derived_probe");
-}
-
 
 // ------------------------------------------------------------------------
 // Members of derived_probe_builder
@@ -166,6 +146,15 @@ derived_probe_builder::get_param (std::map<std::string, literal*> const & params
 }
 
 
+bool
+derived_probe_builder::has_null_param (std::map<std::string, literal*> const & params,
+                                       const std::string& key)
+{
+  map<string, literal *>::const_iterator i = params.find(key);
+  return (i != params.end() && i->second == NULL);
+}
+
+
 
 // ------------------------------------------------------------------------
 // Members of match_key.
@@ -382,6 +371,15 @@ match_node::find_and_build (systemtap_session& s,
 }
 
 
+void
+match_node::build_no_more (systemtap_session& s)
+{
+  for (sub_map_iterator_t i = sub.begin(); i != sub.end(); i++)
+    i->second->build_no_more (s);
+  if (end) end->build_no_more (s);
+}
+
+
 // ------------------------------------------------------------------------
 // Alias probes
 // ------------------------------------------------------------------------
@@ -390,14 +388,13 @@ struct alias_derived_probe: public derived_probe
 {
   alias_derived_probe (probe* base): derived_probe (base) {}
 
-  void register_probe (systemtap_session& s) { }
+  void upchuck () { throw semantic_error ("inappropriate", this->tok); }
 
-  // alias probes should be ultimately expanded to other derived_probe
-  // types, and not themselves emitted.
-  void emit_registrations_start (translator_output* o, unsigned index) { throw semantic_error ("inappropriate", this->tok); }
-  void emit_registrations_end (translator_output* o, unsigned index) { throw semantic_error ("inappropriate", this->tok); }
-  void emit_deregistrations (translator_output* o) { throw semantic_error ("inappropriate", this->tok); }
-  void emit_probe_entries (translator_output* o) { throw semantic_error ("inappropriate", this->tok); }
+  // Alias probes are immediately expanded to other derived_probe
+  // types, and are not themselves emitted or listed in
+  // systemtap_session.probes
+
+  void join_group (systemtap_session&) { upchuck (); }
 };
 
 
@@ -414,7 +411,7 @@ alias_expansion_builder
   virtual void build(systemtap_session & sess,
 		     probe * use, 
 		     probe_point * location,
-		     std::map<std::string, literal *> const & parameters,
+		     std::map<std::string, literal *> const &,
 		     vector<derived_probe *> & finished_results)
   {
     // We're going to build a new probe and wrap it up in an
@@ -506,10 +503,14 @@ systemtap_session::register_library_aliases()
             }
           catch (const semantic_error& e)
             {
-              cerr << "while: registering probe alias ";
-              alias->printsig(cerr);
-              cerr << endl;
-              print_error (e);
+              semantic_error* er = new semantic_error (e); // copy it
+              stringstream msg;
+              msg << e.msg2;
+              msg << " while registering probe alias ";
+              alias->printsig(msg);
+              er->msg2 = msg.str();
+              print_error (* er);
+              delete er;
             }
 	}
     }
@@ -565,8 +566,14 @@ derive_probes (systemtap_session& s,
       catch (const semantic_error& e)
         {
           // XXX: prefer not to print_error at every nest/unroll level
-          s.print_error (e);
-          cerr << "while: resolving probe point " << *loc << endl;
+
+          semantic_error* er = new semantic_error (e); // copy it
+          stringstream msg;
+          msg << e.msg2;
+          msg << " while resolving probe point " << *loc;
+          er->msg2 = msg.str();
+          s.print_error (* er);
+          delete er;
         }
 
       loc->optional = old_loc_opt;
@@ -766,7 +773,7 @@ semantic_pass_vars (systemtap_session & sess)
 	sess.probes[i]->body->visit (&chk);
     }  
 
-  return sess.num_errors;
+  return sess.num_errors();
 }
 
 // ------------------------------------------------------------------------
@@ -868,7 +875,7 @@ semantic_pass_stats (systemtap_session & sess)
 	}
     }
   
-  return sess.num_errors;
+  return sess.num_errors();
 }
 
 // ------------------------------------------------------------------------
@@ -945,7 +952,8 @@ semantic_pass_symbols (systemtap_session& s)
           for (unsigned j=0; j<dps.size(); j++)
             {
               derived_probe* dp = dps[j];
-	      dp->register_probe (s);
+              s.probes.push_back (dp);
+              dp->join_group (s);
 
               try 
                 {
@@ -960,8 +968,12 @@ semantic_pass_symbols (systemtap_session& s)
             }
         }
     }
+
+  // Inform all derived_probe builders that we're done with
+  // all resolution, so it's time to release caches.
+  s.pattern_root->build_no_more (s);
   
-  return s.num_errors; // all those print_error calls
+  return s.num_errors(); // all those print_error calls
 }
 
 
@@ -996,11 +1008,18 @@ semantic_pass (systemtap_session& s)
 
 
 systemtap_session::systemtap_session ():
+  // NB: pointer members must be manually initialized!
   pattern_root(new match_node),
-  user_file (0), op (0), up (0),
-  kprobes_text_initialized (false),
-  kprobes_text_start (0), kprobes_text_end (0),
-  num_errors (0)
+  user_file (0),
+  be_derived_probes(0), 
+  dwarf_derived_probes(0), 
+  uprobe_derived_probes(0), 
+  timer_derived_probes(0), 
+  profile_derived_probes(0), 
+  mark_derived_probes(0), 
+  hrtimer_derived_probes(0), 
+  perfmon_derived_probes(0), 
+  op (0), up (0)
 {
 }
 
@@ -1008,14 +1027,27 @@ systemtap_session::systemtap_session ():
 void
 systemtap_session::print_error (const semantic_error& e)
 {
-  cerr << "semantic error: " << e.what ();
+  string message_str;
+  stringstream message;
+
+  message << "semantic error: " << e.what ();
   if (e.tok1 || e.tok2)
-    cerr << ": ";
-  if (e.tok1) cerr << *e.tok1;
-  cerr << e.msg2;
-  if (e.tok2) cerr << *e.tok2;
-  cerr << endl;
-  num_errors ++;
+    message << ": ";
+  if (e.tok1) message << *e.tok1;
+  message << e.msg2;
+  if (e.tok2) message << *e.tok2;
+  message << endl;
+  message_str = message.str();
+
+  // Duplicate elimination
+  if (seen_errors.find (message_str) == seen_errors.end())
+    {
+      seen_errors.insert (message_str);
+      cerr << message_str;
+    }
+
+  if (e.chain)
+    print_error (* e.chain);
 }
 
 
@@ -1065,9 +1097,10 @@ symresolution_info::visit_foreach_loop (foreach_loop* e)
 	    array->referent = d;
 	  else
             {
-              cerr << "while searching for arity " << e->indexes.size()
-                   << " array:\n";
-              throw semantic_error ("unresolved global array " + array->name, e->tok);
+              stringstream msg;
+              msg << "unresolved arity-" << e->indexes.size()
+                  << " global array " << array->name;
+              throw semantic_error (msg.str(), e->tok);
             }
 	}
     }
@@ -1077,6 +1110,9 @@ symresolution_info::visit_foreach_loop (foreach_loop* e)
       hist->visit (this);
     }
 
+  if (e->limit)
+    e->limit->visit (this);
+
   e->block->visit (this);
 }
 
@@ -1205,8 +1241,10 @@ symresolution_info::visit_functioncall (functioncall* e)
     e->referent = d;
   else
     {
-      cerr << "while searching for arity " << e->args.size() << " function:\n";
-      throw semantic_error ("unresolved function call", e->tok);
+      stringstream msg;
+      msg << "unresolved arity-" << e->args.size()
+          << " function";
+      throw semantic_error (msg.str(), e->tok);
     }
 }
 
@@ -1459,11 +1497,22 @@ dead_assignment_remover::visit_assignment (assignment* e)
       // clog << "Checking assignment to " << leftvar->name << " at " << *e->tok << endl;
       if (vut.read.find(leftvar) == vut.read.end()) // var never read?
         {
-          if (session.verbose>2)
-            clog << "Eliding assignment to " << leftvar->name 
-                 << " at " << *e->tok << endl;
-          *current_expr = e->right; // goodbye assignment*
-          relaxed_p = false;
+          // NB: Not so fast!  The left side could be an array whose
+          // index expressions may have side-effects.  This would be
+          // OK if we could replace the array assignment with a 
+          // statement-expression containing all the index expressions
+          // and the rvalue... but we can't.
+
+          varuse_collecting_visitor vut;
+          e->left->visit (& vut);
+          if (vut.side_effect_free ()) // XXX: use _wrt() once we track focal_vars
+            {
+              if (session.verbose>2)
+                clog << "Eliding assignment to " << leftvar->name 
+                     << " at " << *e->tok << endl;
+              *current_expr = e->right; // goodbye assignment*
+              relaxed_p = false;
+            }
         }
     }
 }
@@ -1543,23 +1592,7 @@ dead_stmtexpr_remover::visit_expr_statement (expr_statement *s)
 
   varuse_collecting_visitor vut;
   s->value->visit (& vut);
-
-  // Note that side-effect-freeness is not simply this test:
-  //
-  // (vut.written.empty() && !vut.embedded_seen)
-  //
-  // That's because the vut.written list may consist of local
-  // variables of called functions.  Visible side-effects occur if
-  // *our* locals, or any *globals* are written-to.
-
-
-  set<vardecl*> intersection;
-  insert_iterator<set<vardecl*> > int_it (intersection, intersection.begin());
-  set_intersection (vut.written.begin(), vut.written.end(),
-                    focal_vars.begin(), focal_vars.end(),
-                    int_it);
-
-  if (intersection.empty() && ! vut.embedded_seen)
+  if (vut.side_effect_free_wrt (focal_vars))
     {
       if (session.verbose>2)
         clog << "Eliding side-effect-free expression "
@@ -1575,15 +1608,6 @@ dead_stmtexpr_remover::visit_expr_statement (expr_statement *s)
       
       relaxed_p = false;
     }
-  else if(session.verbose>3)
-    {
-      clog << "keeping expression " << *s->tok 
-           << " because it writes: ";
-      for (std::set<vardecl*>::iterator k = intersection.begin();
-           k != intersection.end(); k++)
-        clog << (*k)->name << " ";
-      clog << "and/or embedded: " << vut.embedded_seen << endl;
-    }  
 }
 
 
@@ -1828,18 +1852,20 @@ semantic_pass_types (systemtap_session& s)
         }
       
       if (ti.num_newly_resolved == 0) // converged
-        if (ti.num_still_unresolved == 0)
-          break; // successfully
-        else if (! ti.assert_resolvability)
-          ti.assert_resolvability = true; // last pass, with error msgs
-        else
-          { // unsuccessful conclusion
-            rc ++;
-            break;
-          }
+        {
+          if (ti.num_still_unresolved == 0)
+            break; // successfully
+          else if (! ti.assert_resolvability)
+            ti.assert_resolvability = true; // last pass, with error msgs
+          else
+            { // unsuccessful conclusion
+              rc ++;
+              break;
+            }
+        }
     }
   
-  return rc + s.num_errors;
+  return rc + s.num_errors();
 }
 
 
@@ -2320,7 +2346,7 @@ typeresolution_info::visit_block (block* e)
 
 
 void
-typeresolution_info::visit_embeddedcode (embeddedcode* e)
+typeresolution_info::visit_embeddedcode (embeddedcode*)
 {
 }
 
@@ -2412,13 +2438,19 @@ typeresolution_info::visit_foreach_loop (foreach_loop* e)
 	}
     }
 
+  if (e->limit)
+    {
+      t = pe_long;
+      e->limit->visit (this);
+    }
+
   t = pe_unknown;
   e->block->visit (this);  
 }
 
 
 void
-typeresolution_info::visit_null_statement (null_statement* e)
+typeresolution_info::visit_null_statement (null_statement*)
 {
 }
 
@@ -2464,19 +2496,19 @@ typeresolution_info::visit_delete_statement (delete_statement* e)
 
 
 void
-typeresolution_info::visit_next_statement (next_statement* s)
+typeresolution_info::visit_next_statement (next_statement*)
 {
 }
 
 
 void
-typeresolution_info::visit_break_statement (break_statement* s)
+typeresolution_info::visit_break_statement (break_statement*)
 {
 }
 
 
 void
-typeresolution_info::visit_continue_statement (continue_statement* s)
+typeresolution_info::visit_continue_statement (continue_statement*)
 {
 }
 
@@ -2698,11 +2730,18 @@ typeresolution_info::invalid (const token* tok, exp_type pe)
       string nm = (current_function ? current_function->name :
                    current_probe ? current_probe->name :
                    "?");
-      cerr << "semantic error: " + nm + " with invalid type " << pe << " for ";
-      if (tok)
-        cerr << *tok;
+      if (tok && tok->type == tok_operator)
+        {
+	  cerr << "semantic error: " + nm + " uses invalid " << *tok;
+        }
       else
-        cerr << "a token";
+        {
+	  cerr << "semantic error: " + nm + " with invalid type " << pe << " for ";
+	  if (tok)
+	      cerr << *tok;
+	  else
+	      cerr << "a token";
+	}
       cerr << endl;
     }
 }
@@ -2729,7 +2768,7 @@ typeresolution_info::mismatch (const token* tok, exp_type t1, exp_type t2)
 
 
 void
-typeresolution_info::resolved (const token* tok, exp_type t)
+typeresolution_info::resolved (const token*, exp_type)
 {
   num_newly_resolved ++;
   // cerr << "resolved " << *e->tok << " type " << t << endl;
