@@ -1,5 +1,5 @@
 // -*- C++ -*-
-// Copyright (C) 2005, 2006 Red Hat Inc.
+// Copyright (C) 2005-2007 Red Hat Inc.
 //
 // This file is part of systemtap, and is free software.  You can
 // redistribute it and/or modify it under the terms of the GNU General
@@ -115,16 +115,25 @@ struct derived_probe: public probe
   virtual probe* basest () { return base->basest(); }
   virtual ~derived_probe () {}
   virtual void join_group (systemtap_session& s) = 0;
-  virtual probe_point* sole_location ();
+  virtual probe_point* sole_location () const;
+  virtual void printsig (std::ostream &o) const;
+  void printsig_nested (std::ostream &o) const;
 
-  virtual void emit_probe_context_vars (translator_output* o) {}
+  virtual void emit_probe_context_vars (translator_output*) {}
   // From within unparser::emit_common_header, add any extra variables
   // to this probe's context locals.
 
+  virtual void initialize_probe_context_vars (translator_output*) {}
+  // From within unparser::emit_probe, initialized any extra variables
+  // in this probe's context locals.
+
 public:
   static void emit_common_header (translator_output* o);
   // from c_unparser::emit_common_header
   // XXX: probably can move this stuff to a probe_group::emit_module_decls
+
+  virtual bool needs_global_locks () { return true; }
+  // by default, probes need locks around global variables
 };
 
 // ------------------------------------------------------------------------
@@ -170,7 +179,10 @@ struct derived_probe_builder
 		     std::map<std::string, literal*> const & parameters,
 		     std::vector<derived_probe*> & finished_results) = 0;
   virtual ~derived_probe_builder() {}
+  virtual void build_no_more (systemtap_session &) {}
 
+  static bool has_null_param (std::map<std::string, literal*> const & parameters,
+                              const std::string& key);
   static bool get_param (std::map<std::string, literal*> const & parameters,
                          const std::string& key, std::string& value);
   static bool get_param (std::map<std::string, literal*> const & parameters,
@@ -210,6 +222,7 @@ match_node
   void find_and_build (systemtap_session& s,
                        probe* p, probe_point *loc, unsigned pos,
                        std::vector<derived_probe *>& results);
+  void build_no_more (systemtap_session &s);
 
   match_node* bind(match_key const & k);
   match_node* bind(std::string const & k);
