@@ -1017,24 +1017,25 @@ c_translate_location (struct obstack *pool,
 /* Emit "uintNN_t TARGET = ...;".  */
 static bool
 emit_base_fetch (struct obstack *pool, Dwarf_Word byte_size,
-		 const char *target, bool decl, struct location *loc)
+                 bool signed_p, const char *target, struct location *loc)
 {
-  if (decl)
-    obstack_printf (pool, "uint%" PRIu64 "_t ", byte_size * 8);
+  obstack_printf (pool, "%s = ", target);
+
+  /* Emit size/signed coercion. */ 
+  obstack_printf (pool, "(%sint%" PRIu64 "_t)", 
+                  (signed_p ? "" : "u"), byte_size * 8);
 
   switch (loc->type)
     {
     case loc_address:
       if (byte_size != 0 && byte_size != (Dwarf_Word) -1)
-	obstack_printf (pool, "%s = deref (%" PRIu64 ", addr);",
-			target, byte_size);
+	obstack_printf (pool, "deref (%" PRIu64 ", addr);", byte_size);
       else
-	obstack_printf (pool, "%s = deref (sizeof %s, addr);",
-			target, target);
+	obstack_printf (pool, "deref (sizeof %s, addr);", target);
       return true;
 
     case loc_register:
-      obstack_printf (pool, "%s = fetch_register (%u);", target, loc->regno);
+      obstack_printf (pool, "fetch_register (%u);", loc->regno);
       break;
 
     case loc_noncontiguous:
@@ -1092,7 +1093,8 @@ discontiguify (struct obstack *pool, int indent, struct location *loc,
     {
       if (loc->type != loc_noncontiguous)
 	return (loc->byte_size ?: total_bytes) <= max_piece_bytes;
-      for (struct location *p = loc->pieces; p != NULL; p = p->next)
+      struct location *p;
+      for (p = loc->pieces; p != NULL; p = p->next)
 	if (p->byte_size > max_piece_bytes)
 	  return false;
       return true;
@@ -1188,7 +1190,8 @@ declare_noncontig_union (struct obstack *pool, int indent,
   obstack_printf (pool, "%*sstruct {\n", indent++ * 2, "");
 
   Dwarf_Word offset = 0;
-  for (struct location *p = loc->pieces; p != NULL; p = p->next)
+  struct location *p;
+  for (p = loc->pieces; p != NULL; p = p->next)
     {
       obstack_printf (pool, "%*suint%" PRIu64 "_t p%" PRIu64 ";\n",
 		      indent * 2, "", p->byte_size * 8, offset);
@@ -1246,7 +1249,7 @@ get_bitfield (struct location *loc,
    at the *INPUT location and store it in lvalue TARGET.  */
 static void
 translate_base_fetch (struct obstack *pool, int indent, Dwarf_Word byte_size,
-		      struct location **input, const char *target)
+                      bool signed_p, struct location **input, const char *target)
 {
   bool deref = false;
 
@@ -1267,7 +1270,8 @@ translate_base_fetch (struct obstack *pool, int indent, Dwarf_Word byte_size,
 	  *input = newp;
 
 	  snprintf (&piece[sizeof "u.pieces.p" - 1], 20, "%" PRIu64, offset);
-	  translate_base_fetch (pool, indent, p->byte_size, input, piece);
+	  translate_base_fetch (pool, indent, p->byte_size, signed_p /* ? */,
+                                input, piece);
 	  (*input)->type = loc_fragment;
 
 	  offset += p->byte_size;
@@ -1285,7 +1289,7 @@ translate_base_fetch (struct obstack *pool, int indent, Dwarf_Word byte_size,
       case 4:
       case 8:
 	obstack_printf (pool, "%*s", indent * 2, "");
-	deref = emit_base_fetch (pool, byte_size, target, false, *input);
+	deref = emit_base_fetch (pool, byte_size, signed_p, target, *input);
 	obstack_printf (pool, "\n");
 	break;
 
@@ -1332,6 +1336,13 @@ c_translate_fetch (struct obstack *pool, int indent,
       || dwarf_formudata (&size_attr, &byte_size) != 0)
     byte_size = base_byte_size (typedie, *input);
 
+  Dwarf_Attribute encoding_attr;
+  Dwarf_Word encoding;
+  if (dwarf_attr_integrate (die, DW_AT_encoding, &encoding_attr) == NULL
+      || dwarf_formudata (&encoding_attr, &encoding) != 0)
+    encoding = DW_ATE_unsigned; /* default */
+  bool signed_p = (encoding == DW_ATE_signed);
+
   *input = discontiguify (pool, indent, *input, byte_size,
 			  max_fetch_size (*input, die));
 
@@ -1340,7 +1351,7 @@ c_translate_fetch (struct obstack *pool, int indent,
       /* This is a bit field.  Fetch the containing base type into a
 	 temporary variable.  */
 
-      translate_base_fetch (pool, indent, byte_size, input, "tmp");
+      translate_base_fetch (pool, indent, byte_size, signed_p, input, "tmp");
       (*input)->type = loc_fragment;
       (*input)->address.declare = "tmp";
 
@@ -1357,7 +1368,7 @@ c_translate_fetch (struct obstack *pool, int indent,
       *input = loc;
     }
   else
-    translate_base_fetch (pool, indent, byte_size, input, target);
+    translate_base_fetch (pool, indent, byte_size, signed_p, input, target);
 }
 
 /* Translate a fragment to store RVALUE into the base-type value of
@@ -1381,8 +1392,9 @@ translate_base_store (struct obstack *pool, int indent, Dwarf_Word byte_size,
 
       Dwarf_Word offset = 0;
       char piece[sizeof "u.pieces.p" + 20] = "u.pieces.p";
-      for (struct location *p = store_loc->pieces; p != NULL; p = p->next)
-	{
+      struct location *p;
+      for (p = store_loc->pieces; p != NULL; p = p->next)
+        {
 	  struct location *newp = obstack_alloc (pool, sizeof *newp);
 	  *newp = *p;
 	  newp->next = NULL;
@@ -1442,6 +1454,13 @@ c_translate_store (struct obstack *pool, int indent,
       || dwarf_formudata (&size_attr, &byte_size) != 0)
     byte_size = base_byte_size (typedie, *input);
 
+  Dwarf_Attribute encoding_attr;
+  Dwarf_Word encoding;
+  if (dwarf_attr_integrate (die, DW_AT_encoding, &encoding_attr) == NULL
+      || dwarf_formudata (&encoding_attr, &encoding) != 0)
+    encoding = DW_ATE_unsigned; /* default */
+  bool signed_p = (encoding == DW_ATE_signed);
+
   *input = discontiguify (pool, indent, *input, byte_size,
 			  max_fetch_size (*input, die));
 
@@ -1452,7 +1471,7 @@ c_translate_store (struct obstack *pool, int indent,
       /* This is a bit field.  Fetch the containing base type into a
 	 temporary variable.  */
 
-      translate_base_fetch (pool, indent, byte_size, input, "tmp");
+      translate_base_fetch (pool, indent, byte_size, signed_p, input, "tmp");
       (*input)->type = loc_fragment;
       (*input)->address.declare = "tmp";
 
@@ -1499,7 +1518,14 @@ c_translate_pointer (struct obstack *pool, int indent,
 	  dwarf_diename (typedie) ?: "<anonymous>",
 	  dwarf_errmsg (-1));
 
-  translate_base_fetch (pool, indent + 1, byte_size, input, "addr");
+  Dwarf_Attribute encoding_attr;
+  Dwarf_Word encoding;
+  if (dwarf_attr_integrate (typedie, DW_AT_encoding, &encoding_attr) == NULL
+      || dwarf_formudata (&encoding_attr, &encoding) != 0)
+    encoding = DW_ATE_unsigned; /* default */
+  bool signed_p = (encoding == DW_ATE_signed);
+
+  translate_base_fetch (pool, indent + 1, byte_size, signed_p, input, "addr");
   (*input)->type = loc_address;
 }
 
@@ -1651,7 +1677,8 @@ emit_header (FILE *out, struct location *loc, unsigned int hindent)
   else
     {
       emit ("%*s{ // DWARF expression:", hindent * 2, "");
-      for (size_t i = 0; i < loc->nops; ++i)
+      size_t i;
+      for (i = 0; i < loc->nops; ++i)
 	{
 	  emit (" %#x", loc->ops[i].atom);
 	  if (loc->ops[i].number2 == 0)
@@ -1692,7 +1719,8 @@ emit_loc_address (FILE *out, struct location *loc, unsigned int indent,
     {
       emit ("%*s{\n", indent * 2, "");
       emit ("%*s%s " STACKFMT, (indent + 1) * 2, "", STACK_TYPE, 0);
-      for (unsigned int i = 1; i < loc->address.stack_depth; ++i)
+      unsigned i;
+      for (i = 1; i < loc->address.stack_depth; ++i)
 	emit (", " STACKFMT, i);
       emit (";\n");
 
@@ -1737,7 +1765,8 @@ c_emit_location (FILE *out, struct location *loc, int indent)
   emit ("%*s{\n", indent * 2, "");
 
   bool declared_addr = false;
-  for (struct location *l = loc; l != NULL; l = l->next)
+  struct location *l;
+  for (l = loc; l != NULL; l = l->next)
     switch (l->type)
       {
       case loc_decl:
