@@ -1,5 +1,5 @@
 // systemtap translator/driver
-// Copyright (C) 2005-2006 Red Hat Inc.
+// Copyright (C) 2005-2007 Red Hat Inc.
 // Copyright (C) 2005 IBM Corp.
 // Copyright (C) 2006 Intel Corporation.
 //
@@ -31,6 +31,7 @@ extern "C" {
 #include <sys/utsname.h>
 #include <sys/times.h>
 #include <sys/time.h>
+#include <sys/stat.h>
 #include <time.h>
 #include <elfutils/libdwfl.h>
 }
@@ -45,7 +46,7 @@ version ()
     << "SystemTap translator/driver "
     << "(version " << VERSION << " built " << DATE << ")" << endl
     << "(Using " << dwfl_version (NULL) << " libraries.)" << endl
-    << "Copyright (C) 2005-2006 Red Hat, Inc. and others" << endl
+    << "Copyright (C) 2005-2007 Red Hat, Inc. and others" << endl
     << "This is free software; see the source for copying conditions." << endl;
 }
 
@@ -70,8 +71,9 @@ usage (systemtap_session& s, int exitcode)
     << "   -k         keep temporary directory" << endl
     << "   -u         unoptimized translation" << (s.unoptimized ? " [set]" : "") << endl
     << "   -g         guru mode" << (s.guru_mode ? " [set]" : "") << endl
-    << "   -b         bulk (relayfs) mode" << (s.bulk_mode ? " [set]" : "") << endl
-    << "   -M         Don't merge per-cpu files for bulk (relayfs) mode" << (s.merge ? "" : " [set]") << endl
+    << "   -P         prologue-searching for function probes" 
+    << (s.prologue_searching ? " [set]" : "") << endl
+    << "   -b         bulk (percpu file) mode" << (s.bulk_mode ? " [set]" : "") << endl
     << "   -s NUM     buffer size in megabytes, instead of "
     << s.buffer_size << endl
     << "   -p NUM     stop after pass NUM 1-5, instead of "
@@ -88,7 +90,7 @@ usage (systemtap_session& s, int exitcode)
     << "   -D NM=VAL  emit macro definition into generated C code" << endl
     << "   -R DIR     look in DIR for runtime, instead of" << endl
     <<      "              " << s.runtime_path << endl
-    << "   -r RELEASE use kernel RELEASE, instead of "
+    << "   -r RELEASE cross-compile to kernel RELEASE, instead of "
     << s.kernel_release << endl
     << "   -m MODULE  set probe module name, instead of "
     << s.module_name << endl
@@ -96,7 +98,7 @@ usage (systemtap_session& s, int exitcode)
     << "   -c CMD     start the probes, run CMD, and exit when it finishes"
     << endl
     << "   -x PID     sets target() to PID" << endl
-    << "   -t         benchmarking timing information generated" << endl
+    << "   -t         collect probe timing information" << endl
     ;
   // -d: dump safety-related external references
 
@@ -104,27 +106,29 @@ usage (systemtap_session& s, int exitcode)
 }
 
 
-// little utility function
-
-template <typename T>
-static string
-stringify(T t)
-{
-  ostringstream s;
-  s << t;
-  return s.str ();
-}
-
-
 static void
 printscript(systemtap_session& s, ostream& o)
 {
+  if (s.embeds.size() > 0)
+    o << "# global embedded code" << endl;
+  for (unsigned i=0; i<s.embeds.size(); i++)
+    {
+      embeddedcode* ec = s.embeds[i];
+      ec->print (o);
+      o << endl;
+    }
+
   if (s.globals.size() > 0)
     o << "# globals" << endl;
   for (unsigned i=0; i<s.globals.size(); i++)
     {
       vardecl* v = s.globals[i];
       v->printsig (o);
+      if (s.verbose && v->init)
+        {
+          o << " = ";
+          v->init->print(o);
+        }
       o << endl;
     }
 
@@ -181,6 +185,7 @@ main (int argc, char * const argv [])
   string cmdline_script; // -e PROGRAM
   string script_file; // FILE
   bool have_script = false;
+  bool release_changed = false;
 
   // Initialize defaults
   systemtap_session s;
@@ -193,6 +198,13 @@ main (int argc, char * const argv [])
   s.guru_mode = false;
   s.bulk_mode = false;
   s.unoptimized = false;
+
+#ifdef ENABLE_PROLOGUES
+  s.prologue_searching = true;
+#else
+  s.prologue_searching = false;
+#endif
+
   s.buffer_size = 0;
   s.last_pass = 5;
   s.module_name = "stap_" + stringify(getpid());
@@ -202,6 +214,7 @@ main (int argc, char * const argv [])
   s.target_pid = 0;
   s.merge=true;
   s.perfmon=0;
+  s.symtab = false;
   s.use_cache = true;
 
   const char* s_p = getenv ("SYSTEMTAP_TAPSET");
@@ -251,7 +264,8 @@ main (int argc, char * const argv [])
 
   while (true)
     {
-      int grc = getopt (argc, argv, "hVMvtp:I:e:o:R:r:m:kgc:x:D:bs:u");
+      // NB: also see find_hash(), help(), switch stmt below, stap.1 man page
+      int grc = getopt (argc, argv, "hVMvtp:I:e:o:R:r:m:kgPc:x:D:bs:u");
       if (grc < 0)
         break;
       switch (grc)
@@ -306,12 +320,45 @@ main (int argc, char * const argv [])
 
         case 'm':
           s.module_name = string (optarg);
+	  {
+	    string::size_type len = s.module_name.length();
+
+	    // If the module name ends with '.ko', chop it off since
+	    // modutils doesn't like modules named 'foo.ko.ko'.
+	    if (len > 3 && s.module_name.substr(len - 3, 3) == ".ko")
+	      {
+		s.module_name.erase(len - 3);
+		len -= 3;
+		cerr << "Truncating module name to '" << s.module_name
+		     << "'" << endl;
+	      }
+
+	    // Make sure an empty module name wasn't specified (-m "")
+	    if (len == 0)
+	    {
+		cerr << "Module name cannot be empty." << endl;
+		usage (s, 1);
+	    }
+
+	    // Make sure the module name is only composed of the
+	    // following chars: [_a-zA-Z0-9]
+	    const string identchars("_" "abcdefghijklmnopqrstuvwxyz"
+				    "ABCDEFGHIJKLMNOPQRSTUVWXYZ" "0123456789");
+	    if (s.module_name.find_first_not_of(identchars) != string::npos)
+	      {
+		cerr << "Invalid module name (must only be composed of"
+		    " characters [_a-zA-Z0-9])." << endl;
+		usage (s, 1);
+	      }
+	  }
+
 	  cerr << "Warning: using '-m' disables cache support." << endl;
 	  s.use_cache = false;
           break;
 
         case 'r':
           s.kernel_release = string (optarg);
+	  release_changed = true;
           break;
 
         case 'k':
@@ -322,6 +369,10 @@ main (int argc, char * const argv [])
           s.guru_mode = true;
           break;
 
+        case 'P':
+          s.prologue_searching = true;
+          break;
+
         case 'b':
           s.bulk_mode = true;
           break;
@@ -373,6 +424,13 @@ main (int argc, char * const argv [])
       usage (s, 1);
     }
 
+  if (s.last_pass > 4 && release_changed)
+    {
+      if (s.verbose)
+        cerr << "Warning: changing last pass to 4 since cross-compiling" << endl;
+      s.last_pass = 4;
+    }
+
   for (int i = optind; i < argc; i++)
     {
       if (! have_script)
@@ -405,8 +463,9 @@ main (int argc, char * const argv [])
 
   s.kernel_base_release.assign(s.kernel_release, 0, s.kernel_release.find('-'));
 
-
   // arguments parsed; get down to business
+  if (s.verbose > 1)
+    version ();
 
   // Create a temporary directory to build within.
   // Be careful with this, as "s.tmpdir" is "rm -rf"'d at the end.
@@ -441,12 +500,20 @@ main (int argc, char * const argv [])
   gettimeofday (&tv_before, NULL);
 
   // PASS 1a: PARSING USER SCRIPT
-  // XXX: pass args vector, so parser (or lexer?) can substitute
-  // $1..$NN with actual arguments
+
+  struct stat user_file_stat;
+  int user_file_stat_rc = -1;
+
   if (script_file == "-")
-    s.user_file = parser::parse (s, cin, s.guru_mode);
+    {
+      s.user_file = parser::parse (s, cin, s.guru_mode);
+      user_file_stat_rc = fstat (STDIN_FILENO, & user_file_stat);
+    }
   else if (script_file != "")
-    s.user_file = parser::parse (s, script_file, s.guru_mode);
+    {
+      s.user_file = parser::parse (s, script_file, s.guru_mode);
+      user_file_stat_rc = stat (script_file.c_str(), & user_file_stat);
+    }
   else
     {
       istringstream ii (cmdline_script);
@@ -503,11 +570,24 @@ main (int argc, char * const argv [])
           for (unsigned j=0; j<globbuf.gl_pathc; j++)
             {
               // privilege only for /usr/share/systemtap?
+              
               stapfile* f = parser::parse (s, globbuf.gl_pathv[j], true);
               if (f == 0)
                 rc ++;
               else
                 s.library_files.push_back (f);
+
+              struct stat tapset_file_stat;
+              int stat_rc = stat (globbuf.gl_pathv[j], & tapset_file_stat);
+              if (stat_rc == 0 && user_file_stat_rc == 0 &&
+                  user_file_stat.st_dev == tapset_file_stat.st_dev &&
+                  user_file_stat.st_ino == tapset_file_stat.st_ino)
+                {
+                  clog << "usage error: tapset file '" << globbuf.gl_pathv[j]
+                       << "' cannot be run directly as a session script." << endl;
+                  rc ++;
+                }
+
             }
 
           globfree (& globbuf);
@@ -573,6 +653,7 @@ main (int argc, char * const argv [])
   if (s.verbose) clog << "Pass 2: analyzed script: "
                       << s.probes.size() << " probe(s), "
                       << s.functions.size() << " function(s), "
+                      << s.embeds.size() << " embed(s), "
                       << s.globals.size() << " global(s) in "
                       << TIMESPRINT
                       << endl;
@@ -587,17 +668,15 @@ main (int argc, char * const argv [])
     {
       ostringstream o;
       unsigned saved_verbose;
-	  
-      // Make sure we're in verbose mode, so that printscript()
-      // will output function/probe bodies.
-      saved_verbose = s.verbose;
-      s.verbose = 3;
 
-      // Print script to 'o'
-      printscript(s, o);
-
-      // Restore original verbose mode setting.
-      s.verbose = saved_verbose;
+      {
+        // Make sure we're in verbose mode, so that printscript()
+        // will output function/probe bodies.
+        saved_verbose = s.verbose;
+        s.verbose = 3;
+        printscript(s, o);  // Print script to 'o'
+        s.verbose = saved_verbose;
+      }
 
       // Generate hash
       find_hash (s, o.str());
@@ -649,6 +728,10 @@ main (int argc, char * const argv [])
   times (& tms_before);
   gettimeofday (&tv_before, NULL);
   rc = compile_pass (s);
+
+  if (rc == 0 && s.last_pass == 4)
+    cout << s.hash_path << endl;
+
   times (& tms_after);
   gettimeofday (&tv_after, NULL);
 
@@ -668,9 +751,9 @@ main (int argc, char * const argv [])
       add_to_cache(s);
     }
 
-  // XXX: what to do if rc==0 && last_pass == 4?  dump .ko file to stdout?
   if (rc || s.last_pass == 4) goto cleanup;
 
+
   // PASS 5: RUN
 pass_5:
   times (& tms_before);
