@@ -1,5 +1,6 @@
 // -*- C++ -*-
-// Copyright (C) 2005, 2006 Red Hat Inc.
+// Copyright (C) 2005-2007 Red Hat Inc.
+// Copyright (C) 2006 Intel Corporation.
 //
 // This file is part of systemtap, and is free software.  You can
 // redistribute it and/or modify it under the terms of the GNU General
@@ -25,17 +26,18 @@ struct token; // parse.h
 struct semantic_error: public std::runtime_error
 {
   const token* tok1;
-  const std::string msg2;
+  std::string msg2;
   const token* tok2;
+  semantic_error *chain;
 
   ~semantic_error () throw () {}
   semantic_error (const std::string& msg):
-    runtime_error (msg), tok1 (0), tok2 (0) {}
+    runtime_error (msg), tok1 (0), tok2 (0), chain (0) {}
   semantic_error (const std::string& msg, const token* t1):
-    runtime_error (msg), tok1 (t1), tok2 (0) {}
+    runtime_error (msg), tok1 (t1), tok2 (0), chain (0) {}
   semantic_error (const std::string& msg, const token* t1,
                   const std::string& m2, const token* t2):
-    runtime_error (msg), tok1 (t1), msg2 (m2), tok2 (t2) {}
+    runtime_error (msg), tok1 (t1), msg2 (m2), tok2 (t2), chain (0) {}
 };
 
 // ------------------------------------------------------------------------
@@ -393,7 +395,9 @@ struct vardecl: public symboldecl
   void set_arity (int arity);
   bool compatible_arity (int a);
   int arity; // -1: unknown; 0: scalar; >0: array
+  int maxsize; // upperbound on size for arrays
   std::vector<exp_type> index_types; // for arrays only
+  literal *init; // for global scalars only
 };
 
 
@@ -464,6 +468,7 @@ struct foreach_loop: public statement
   indexable *base;
   int sort_direction; // -1: decreasing, 0: none, 1: increasing
   unsigned sort_column; // 0: value, 1..N: index
+  expression* limit; // optional iteration limit
 
   statement* block;
   void print (std::ostream& o) const;
@@ -581,6 +586,7 @@ struct probe
   virtual void printsig (std::ostream &o) const;
   virtual probe* basest () { return this; }
   virtual ~probe() {}
+  bool privileged;
 private:
   static unsigned last_probeidx;
 public:
@@ -717,8 +723,10 @@ struct varuse_collecting_visitor: public functioncall_traversing_visitor
   void visit_pre_crement (pre_crement *e);
   void visit_post_crement (post_crement *e);
   void visit_foreach_loop (foreach_loop *s);
-};
 
+  bool side_effect_free ();
+  bool side_effect_free_wrt (const std::set<vardecl*>& vars);
+};
 
 
 
